from __future__ import annotations

from typing import Any, Generic, Iterator, Mapping, OrderedDict, TypeVar, cast

from PyQt5.QtCore import QAbstractListModel, QModelIndex, Qt
from vstools import vs

from ..core import AbstractMainWindow, AudioOutput, QYAMLObject, VideoOutput, VideoOutputNode, main_window, try_load

T = TypeVar('T', VideoOutput, AudioOutput)


class Outputs(Generic[T], QAbstractListModel, QYAMLObject):
    out_type: type[T]
    _items: list[T]

    __slots__ = ('items')

    def __init__(self, main: AbstractMainWindow, local_storage: Mapping[str, T] | None = None) -> None:
        self.setValue(main, local_storage)

    def setValue(self, main: AbstractMainWindow, local_storage: Mapping[str, T] | None = None) -> None:
        super().__init__()
        self.items = list[T]()
        self.main = main

        local_storage, newstorage = (local_storage, False) if local_storage is not None else ({}, True)

        if main.storage_not_found:
            newstorage = False

        outputs = OrderedDict(sorted(vs.get_outputs().items()))

        main.reload_signal.connect(self.clear_outputs)

        for i, vs_output in outputs.items():
            if not isinstance(vs_output, self.vs_type):
                continue
            try:
                output = local_storage[str(i)]
                output.setValue(vs_output, i, newstorage)
            except KeyError:
                output = self.out_type(vs_output, i, newstorage)

            self.items.append(output)

        self._items = list(self.items)

    def clear_outputs(self) -> None:
        for o in self.items:
            o.clear()

    def __getitem__(self, i: int) -> T:
        return self.items[i]

    def __len__(self) -> int:
        return len(self.items)

    def index_of(self, item: T) -> int:
        return self.items.index(item)

    def __iter__(self) -> Iterator[T]:
        return iter(self.items)

    def append(self, item: T) -> int:
        index = len(self.items)
        self.beginInsertRows(QModelIndex(), index, index)
        self.items.append(item)
        self.endInsertRows()

        return len(self.items) - 1

    def clear(self) -> None:
        self.beginRemoveRows(QModelIndex(), 0, len(self.items))
        self.items.clear()
        self.endRemoveRows()

    def data(self, index: QModelIndex, role: int = Qt.UserRole) -> Any:
        if not index.isValid():
            return None
        if index.row() >= len(self.items):
            return None

        if role == Qt.DisplayRole:
            return self.items[index.row()].name
        if role == Qt.EditRole:
            return self.items[index.row()].name
        if role == Qt.UserRole:
            return self.items[index.row()]
        return None

    def rowCount(self, parent: QModelIndex = QModelIndex()) -> int:
        return len(self.items)

    def flags(self, index: QModelIndex) -> Qt.ItemFlags:
        if not index.isValid():
            return cast(Qt.ItemFlags, Qt.ItemIsEnabled)

        return super().flags(index) | Qt.ItemIsEditable

    def setData(self, index: QModelIndex, value: Any, role: int = Qt.EditRole) -> bool:
        if not index.isValid():
            return False
        if not role == Qt.EditRole:
            return False
        if not isinstance(value, str):
            return False

        self.items[index.row()].name = value
        self.dataChanged.emit(index, index, [role])

        return True

    def __getstate__(self) -> Mapping[str, Any]:
        return dict(zip([str(x.index) for x in self.items], self.items), type=self.out_type.__name__)

    def __setstate__(self, state: Mapping[str, T]) -> None:
        type_string = ''
        try_load(state, 'type', str, type_string)

        for key, value in state.items():
            if key == 'type':
                continue
            if not isinstance(key, str):
                raise TypeError(f'Storage loading (Outputs): key {key} is not a string')
            if not isinstance(value, self.out_type):
                raise TypeError(f'Storage loading (Outputs): value of key {key} is not {self.out_type.__name__}')

        self.setValue(main_window(), state)


class VideoOutputs(Outputs[VideoOutput]):
    out_type = VideoOutput
    vs_type = vs.VideoOutputTuple

    _fft_spectr_items = list[VideoOutput]()

    def copy_output_props(self, new: VideoOutput, old: VideoOutput) -> None:
        new.last_showed_frame = old.last_showed_frame
        new.name = old.name

    def get_new_output(self, new_clip: vs.VideoNode, old_output: VideoOutput) -> VideoOutput:
        new_videonode = VideoOutputNode(new_clip, old_output.source.alpha)

        new_output = VideoOutput(new_videonode, old_output.index, False)

        self.copy_output_props(new_output, old_output)

        return new_output

    def switchToNormalView(self) -> None:
        for new, old in zip(self._items, self.items):
            self.copy_output_props(new, old)

        self.items = list(self._items)

    def switchToFFTSpectrumView(self, force_cache: bool = False) -> None:
        try:
            from vsdfft import FFTSpectrum
        except ModuleNotFoundError:
            raise RuntimeError(
                'vspreview: You can\'t chage to this view mode. You\'re missing the `vsdfft` package!'
            )

        if not self._fft_spectr_items or force_cache:
            max_width = max(*(x.width for x in self._items), 140)
            max_height = max(*(x.height for x in self._items), 140)

            self._fft_spectr_items = [
                self.get_new_output(
                    FFTSpectrum(old.source.clip, target_size=(max_width, max_height)), old
                ) for old in self._items
            ]
        else:
            for new, old in zip(self._fft_spectr_items, self.items):
                self.copy_output_props(new, old)

        self.items = self._fft_spectr_items


class AudioOutputs(Outputs[AudioOutput]):
    out_type = AudioOutput
    vs_type = vs.AudioNode



from __future__ import annotations

from typing import Any, Mapping

from ...core import AbstractToolbarSettings, CheckBox, try_load


class CompSettings(AbstractToolbarSettings):
    __slots__ = ('delete_cache_checkbox', )

    DEFAULT_COLLECTION_NAME = ''

    def setup_ui(self) -> None:
        super().setup_ui()

        self.delete_cache_checkbox = CheckBox('Delete images cache after upload')

        self.vlayout.addWidget(self.delete_cache_checkbox)

    def set_defaults(self) -> None:
        self.delete_cache_checkbox.setChecked(True)

    @property
    def delete_cache_enabled(self) -> bool:
        return self.delete_cache_checkbox.isChecked()

    def __getstate__(self) -> Mapping[str, Any]:
        return {
            'delete_cache_enabled': self.delete_cache_enabled
        }

    def __setstate__(self, state: Mapping[str, Any]) -> None:
        try_load(state, 'delete_cache_enabled', bool, self.delete_cache_checkbox.setChecked)
        super().__setstate__(state)



from __future__ import annotations

import logging

from ...core import AbstractMainWindow, AbstractToolbar, LineEdit, PushButton
from ...utils import debug, vs_clear_cache
from .settings import DebugSettings


class DebugToolbar(AbstractToolbar):
    _no_visibility_choice = True

    __slots__ = ('exec_lineedit', )

    def __init__(self, main: AbstractMainWindow) -> None:
        super().__init__(main, DebugSettings())

        self.setup_ui()

        if self.settings.DEBUG_TOOLBAR_BUTTONS_PRINT_STATE:
            self.filter = debug.EventFilter(main)
            self.main.toolbars.main.widget.installEventFilter(self.filter)
            for toolbar in self.main.toolbars:
                toolbar.widget.installEventFilter(self.filter)

        self.set_qobject_names()

    def setup_ui(self) -> None:
        super().setup_ui()

        self.exec_lineedit = LineEdit(
            self, editingFinished=self.exec_button_clicked,
            placeholderText='Python statement in context of DebugToolbar.exec_button_clicked()'
        )

        self.hlayout.addWidgets([
            PushButton('Test', self, clicked=self.test_button_clicked),
            PushButton('Break', self),
            self.exec_lineedit,
            PushButton('Exec', self, clicked=self.exec_button_clicked)
        ])

        self.hlayout.addStretch()

    def test_button_clicked(self, checked: bool | None = None) -> None:
        vs_clear_cache()

    def exec_button_clicked(self, checked: bool | None = None) -> None:
        try:
            exec(self.exec_lineedit.text())
        except BaseException as e:
            logging.error(e)

    def break_button_clicked(self, checked: bool | None = None) -> None:
        breakpoint()



from __future__ import annotations

from functools import partial

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QColor, QPainter, QPaintEvent
from PyQt5.QtWidgets import QWidget

from ..abstracts import AbstractMainWindow, Timer
from .graphicsview import DragEventType, GraphicsView


class DragNavigator(QWidget):
    __slots__ = ()

    is_tracking = False
    contentsH = 0
    contentsW = 0
    viewportX = 0
    viewportY = 0
    viewportH = 0
    viewportW = 0

    def __init__(self, main: AbstractMainWindow, graphics_view: GraphicsView) -> None:
        super().__init__(graphics_view)

        self.main = main
        self.graphics_view = graphics_view
        self.graphics_view.dragEvent.connect(self.on_drag)
        self.setAttribute(Qt.WA_NoSystemBackground)
        self.setAttribute(Qt.WA_TransparentForMouseEvents)
        rate = self.main.settings.base_ppi / 96
        self.setGeometry(round(10 * rate), round(10 * rate), round(120 * rate), round(120 * rate))
        self.repaint_timer = Timer(
            timeout=self.repaint_timeout, timerType=Qt.PreciseTimer, interval=self.main.settings.dragnavigator_timeout
        )
        self.graphics_view.verticalScrollBar().valueChanged.connect(partial(self.on_drag, DragEventType.repaint))
        self.graphics_view.horizontalScrollBar().valueChanged.connect(partial(self.on_drag, DragEventType.repaint))

    def on_drag(self, event_type: DragEventType) -> None:
        # while reloading and moving mouse
        if not hasattr(self.main, 'current_output') or not self.main.current_output:
            return

        if event_type == DragEventType.repaint:
            self.repaint_timer.stop()
            self.repaint_timer.start()
        elif event_type == DragEventType.move and not self.is_tracking:
            return
        elif event_type == DragEventType.start:
            self.is_tracking = True
        elif event_type == DragEventType.stop:
            self.is_tracking = False
            self.setVisible(False)
            return

        scrollbarW = self.graphics_view.horizontalScrollBar()
        scrollbarH = self.graphics_view.verticalScrollBar()

        if not any({scrollbarW.isVisible(), scrollbarH.isVisible()}):
            return self.hide()

        self.setVisible(True)
        self.draw(
            self.main.current_output.width,
            self.main.current_output.height,
            int(scrollbarW.value() / self.graphics_view.currentZoom),
            int(scrollbarH.value() / self.graphics_view.currentZoom),
            int(self.graphics_view.width() / self.graphics_view.currentZoom),
            int(self.graphics_view.height() / self.graphics_view.currentZoom)
        )

    def draw(
        self, contentsW: int, contentsH: int, viewportX: int, viewportY: int, viewportW: int, viewportH: int
    ) -> None:
        self.contentsW = contentsW
        self.contentsH = contentsH
        self.viewportW = min(viewportW, contentsW)
        self.viewportH = min(viewportH, contentsH)
        self.viewportX = min(viewportX, contentsW)
        self.viewportY = min(viewportY, contentsH)

        self.repaint()

    def repaint_timeout(self) -> None:
        self.hide()
        self.repaint_timer.stop()

    def paintEvent(self, event: QPaintEvent) -> None:
        if(
            (self.contentsW == 0) or (self.contentsH == 0) or
            (self.viewportW == 0) or (self.viewportH == 0) or
            (self.viewportX >= self.contentsW) or (self.viewportY >= self.contentsH)
        ):
            event.ignore()
            return

        norm = 100.0 / max(self.contentsW, self.contentsH, self.viewportW, self.viewportH)

        normContentsWidth, normContentsHeight = round(self.contentsW * norm), round(self.contentsH * norm)
        normViewportWidth, normViewportHeight = round(self.viewportW * norm), round(self.viewportH * norm)
        normVieportX = min(round(self.viewportX * norm), normContentsWidth - normViewportWidth)
        normViwportY = min(round(self.viewportY * norm), normContentsHeight - normViewportHeight)

        cX1 = cY1 = 0
        cX2, cY2 = normContentsWidth - 1, normContentsHeight - 1

        vX1, vY1 = normVieportX, normViwportY
        vX2, vY2 = normVieportX + normViewportWidth - 1, normViwportY + normViewportHeight - 1

        painter = QPainter(self)

        painter.setPen(QColor(255, 0, 255))
        painter.drawLine(cX1, cY1, cX2, cY1)
        painter.drawLine(cX2, cY1, cX2, cY2)
        painter.drawLine(cX2, cY2, cX1, cY2)
        painter.drawLine(cX1, cY2, cX1, cY1)

        painter.setPen(QColor(0, 255, 0))
        painter.drawLine(vX1, vY1, vX2, vY1)
        painter.drawLine(vX2, vY1, vX2, vY2)
        painter.drawLine(vX2, vY2, vX1, vY2)
        painter.drawLine(vX1, vY2, vX1, vY1)

        event.accept()



from __future__ import annotations

from typing import Any, Iterator, cast

from PyQt5.QtCore import QAbstractListModel, QModelIndex, Qt

from ..core import PictureType


class PictureTypes(QAbstractListModel):
    __slots__ = ('items',)

    def __init__(self) -> None:
        super().__init__()
        self.items = PictureType.list()

    def __getitem__(self, i: int) -> PictureType:
        return self.items[i]

    def __len__(self) -> int:
        return len(self.items)

    def __iter__(self) -> Iterator[PictureType]:
        return iter(self.items)

    def index_of(self, item: PictureType) -> int:
        return self.items.index(item)

    def data(self, index: QModelIndex, role: int = Qt.UserRole) -> Any:
        if (not index.isValid() or index.row() >= len(self.items)):
            return None

        if role == Qt.DisplayRole or role == Qt.EditRole:
            return str(self.items[index.row()])
        if role == Qt.UserRole:
            return self.items[index.row()]
        return None

    def rowCount(self, parent: QModelIndex = QModelIndex()) -> int:
        return len(self.items)

    def flags(self, index: QModelIndex) -> Qt.ItemFlags:
        if not index.isValid():
            return cast(Qt.ItemFlags, Qt.ItemIsEnabled)

        return super().flags(index) | Qt.ItemIsEditable



from . import vsenv  # Load vsenv early!
from . import abstracts, bases, better_abc, types, vsenv  # noqa: F401
from .abstracts import *  # noqa: F401, F403
from .bases import *  # noqa: F401, F403
from .types import *  # noqa: F401, F403



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Generic, TypeVar, cast

from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import QComboBox, QWidget

from ...core import AudioOutput, PictureType, VideoOutput
from ...models import SceningList

ComboBoxT = TypeVar('ComboBoxT', VideoOutput, AudioOutput, SceningList, PictureType, float, str)


class ComboBox(QComboBox, Generic[ComboBoxT]):
    def __class_getitem__(cls, content_type: type[ComboBoxT]) -> type:
        return {
            VideoOutput: _ComboBox_Output,
            AudioOutput: _ComboBox_AudioOutput,
            SceningList: _ComboBox_SceningList,
            PictureType: _ComboBox_PictureType,
            float: _ComboBox_float,
            int: _ComboBox_int,
            str: _ComboBox_string
        }[content_type]

    indexChanged = pyqtSignal(int, int)
    content_type: type[ComboBoxT]

    def __init__(self, parent: QWidget | None = None, **kwargs: Any) -> None:
        super().__init__(parent)

        self.setSizeAdjustPolicy(QComboBox.AdjustToMinimumContentsLengthWithIcon)

        self.oldValue = self.currentData()
        self.oldIndex = self.currentIndex()
        self.currentIndexChanged.connect(self._currentIndexChanged)

        for arg in kwargs:
            getattr(self, 'set' + arg[0].upper() + arg[1:])(kwargs.get(arg))

    def _currentIndexChanged(self, newIndex: int) -> None:
        newValue = self.currentData()
        if newValue is None:
            return
        self.valueChanged.emit(newValue, self.oldValue)
        self.indexChanged.emit(newIndex, self.oldIndex)
        self.oldValue = newValue
        self.oldIndex = newIndex

    def currentValue(self) -> ComboBoxT:
        return cast(ComboBoxT, self.currentData())

    def setCurrentValue(self, newValue: ComboBoxT) -> None:
        i = self.model().index_of(newValue)
        self.setCurrentIndex(i)


class _ComboBox_Output(ComboBox):
    content_type = VideoOutput
    T = VideoOutput
    if TYPE_CHECKING:
        valueChanged = pyqtSignal(content_type, content_type | None)
    else:
        valueChanged = pyqtSignal(content_type, object)


class _ComboBox_AudioOutput(ComboBox):
    content_type = AudioOutput
    T = AudioOutput
    if TYPE_CHECKING:
        valueChanged = pyqtSignal(content_type, content_type | None)
    else:
        valueChanged = pyqtSignal(object, object)


class _ComboBox_SceningList(ComboBox):
    content_type = SceningList
    if TYPE_CHECKING:
        valueChanged = pyqtSignal(content_type, content_type | None)
    else:
        valueChanged = pyqtSignal(content_type, object)


class _ComboBox_PictureType(ComboBox):
    content_type = PictureType
    if TYPE_CHECKING:
        valueChanged = pyqtSignal(content_type, content_type | None)
    else:
        valueChanged = pyqtSignal(content_type, object)


class _ComboBox_float(ComboBox):
    content_type = float
    if TYPE_CHECKING:
        valueChanged = pyqtSignal(content_type, content_type | None)
    else:
        valueChanged = pyqtSignal(content_type, object)


class _ComboBox_int(ComboBox):
    content_type = int
    if TYPE_CHECKING:
        valueChanged = pyqtSignal(content_type, content_type | None)
    else:
        valueChanged = pyqtSignal(content_type, object)


class _ComboBox_string(ComboBox):
    content_type = str
    if TYPE_CHECKING:
        valueChanged = pyqtSignal(content_type, content_type | None)
    else:
        valueChanged = pyqtSignal(content_type, object)




from __future__ import annotations

from fractions import Fraction
import inspect
from pathlib import Path

from vstools import vs

from .timecodes import set_timecodes

__all__ = [
    'set_output'
]


def set_output(
    node: vs.RawNode, name: str | None = None,
    *, timecodes: str | Path | dict[
        tuple[int | None, int | None], float | tuple[int, int] | Fraction
    ] | list[float] | None = None, denominator: int = 1001
) -> None:
    from ..core import main_window

    index = len(vs.get_outputs())

    ref_id = str(id(node))

    if isinstance(node, vs.VideoNode):
        title, node_type = 'Clip', vs.VideoNode
    elif isinstance(node, vs.AudioNode):
        title, node_type = 'Audio', vs.AudioNode
    else:
        title, node_type = 'Node', vs.RawNode

    if not name:
        name = f"{title} {index}"

        current_frame = inspect.currentframe()

        assert current_frame
        assert current_frame.f_back

        for vname, val in reversed(current_frame.f_back.f_locals.items()):
            if (str(id(val)) == ref_id):
                name = vname
                break

    node.set_output(index)
    main_window().set_node_name(node_type, index, name.title())  # type: ignore

    if timecodes:
        set_timecodes(index, timecodes, (
            node.fps_den if (node.fps_den and node.fps_num) else 1001
        ) if denominator is None else denominator)
